@@ -27,6 +27,19 @@
 #include <pango/pangoxft.h>
 #endif
 
+/**
+ * gdk_pango_context_get:
+ * 
+ * Creates a #PangoContext for the default GDK display.
+ *
+ * The context must be freed when you're finished with it.
+ * 
+ * When using GTK+, normally you should use gtk_widget_get_pango_context()
+ * instead of this function, to get the appropriate context for
+ * the widget you intend to render text onto.
+ * 
+ * Return value: a new #PangoContext for the default display
+ **/
 PangoContext *
 gdk_pango_context_get (void)
 {
@@ -36,7 +49,7 @@ gdk_pango_context_get (void)
     {
       const char *val = g_getenv ("GDK_USE_XFT");
 
-      use_xft = val && (atoi (val) != 0);
+      use_xft = val && (atoi (val) != 0) && _gdk_x11_have_render ();
     }
   
   if (use_xft)
