@@ -162,20 +162,22 @@ dates_difference(N_int year1, N_int mm1, N_int dd1,
 
 /*** END OF lib_date routines ********************************************/
 
+/* Spacing around day/week headers and main area, inside those windows */
 #define CALENDAR_MARGIN		 0
-#define CALENDAR_YSEP		 4
-#define CALENDAR_XSEP		 4
+/* Spacing around day/week headers and main area, outside those windows */
 #define INNER_BORDER		 4
+/* Separation between day headers and main area */
+#define CALENDAR_YSEP		 4
+/* Separation between week headers and main area */
+#define CALENDAR_XSEP		 4
 
-#define DAY_XPAD		 2
-#define DAY_YPAD		 2
 #define DAY_XSEP		 0 /* not really good for small calendar */
 #define DAY_YSEP		 0 /* not really good for small calendar */
 
 /* Color usage */
 #define HEADER_FG_COLOR(widget)		 (& (widget)->style->fg[GTK_WIDGET_STATE (widget)])
 #define HEADER_BG_COLOR(widget)		 (& (widget)->style->bg[GTK_WIDGET_STATE (widget)])
-#define DAY_NAME_COLOR(widget)		 (& (widget)->style->bg[GTK_STATE_SELECTED])
+#define DAY_NAME_COLOR(widget)		 (& (widget)->style->base[GTK_STATE_SELECTED])
 #define NORMAL_DAY_COLOR(widget)	 (& (widget)->style->fg[GTK_WIDGET_STATE (widget)])
 #define SELECTION_FOCUS_COLOR(widget)	 (& (widget)->style->fg[GTK_WIDGET_STATE (widget)])
 #define SELECTION_NO_FOCUS_COLOR(widget) (& (widget)->style->mid[GTK_WIDGET_STATE (widget)])
@@ -258,10 +260,9 @@ typedef void (*GtkCalendarSignalDate) (GtkObject *object, guint arg1, guint arg2
 
 static void gtk_calendar_class_init	(GtkCalendarClass *class);
 static void gtk_calendar_init		(GtkCalendar *calendar);
-static void gtk_calendar_destroy	(GtkObject *calendar);
+static void gtk_calendar_finalize	(GObject *calendar);
 static void gtk_calendar_realize	(GtkWidget *widget);
 static void gtk_calendar_unrealize	(GtkWidget *widget);
-static void gtk_calendar_draw_focus	(GtkWidget *widget);
 static void gtk_calendar_size_request	(GtkWidget *widget,
 					 GtkRequisition *requisition);
 static void gtk_calendar_size_allocate	(GtkWidget *widget,
@@ -290,8 +291,6 @@ static void gtk_calendar_paint_week_numbers (GtkWidget *widget);
 static void gtk_calendar_paint_main	    (GtkWidget *widget);
 
 
-static void gtk_calendar_paint		(GtkWidget    *widget,
-					 GdkRectangle *area);
 static void gtk_calendar_paint_arrow	(GtkWidget    *widget,
 					 guint	       arrow);
 static void gtk_calendar_paint_day_num	(GtkWidget    *widget,
@@ -337,15 +336,17 @@ gtk_calendar_get_type (void)
 static void
 gtk_calendar_class_init (GtkCalendarClass *class)
 {
+  GObjectClass   *gobject_class;
   GtkObjectClass *object_class;
   GtkWidgetClass *widget_class;
-  
+
+  gobject_class = (GObjectClass*)  class;
   object_class = (GtkObjectClass*) class;
   widget_class = (GtkWidgetClass*) class;
   
   parent_class = gtk_type_class (GTK_TYPE_WIDGET);
   
-  object_class->destroy = gtk_calendar_destroy;
+  gobject_class->finalize = gtk_calendar_finalize;
 
   widget_class->realize = gtk_calendar_realize;
   widget_class->unrealize = gtk_calendar_unrealize;
@@ -562,7 +563,7 @@ left_x_for_column (GtkCalendar *calendar,
 
   width = GTK_CALENDAR_PRIVATE_DATA (calendar)->day_width;
   if (calendar->display_flags & GTK_CALENDAR_SHOW_WEEK_NUMBERS)
-    x_left =  DAY_XSEP + (width + DAY_XSEP) * column;
+    x_left = CALENDAR_XSEP + (width + DAY_XSEP) * column;
   else
     x_left = CALENDAR_MARGIN + (width + DAY_XSEP) * column;
   
@@ -660,8 +661,8 @@ gtk_calendar_set_month_prev (GtkCalendar *calendar)
 	calendar->selected_day = calendar->selected_day + 1 + month_length[leap (calendar->year)][calendar->month + 1];
       gtk_calendar_select_day (calendar, calendar->selected_day);
     }
-  
-  gtk_calendar_paint (GTK_WIDGET (calendar), NULL);
+
+  gtk_widget_queue_draw (GTK_WIDGET (calendar));
   gtk_calendar_thaw (calendar);
 }
 
@@ -701,8 +702,8 @@ gtk_calendar_set_month_next (GtkCalendar *calendar)
     }
   else
     gtk_calendar_select_day (calendar, calendar->selected_day);
-  
-  gtk_calendar_paint (GTK_WIDGET(calendar), NULL);
+
+  gtk_widget_queue_draw (GTK_WIDGET (calendar));
   gtk_calendar_thaw (calendar);
 }
 
@@ -731,7 +732,7 @@ gtk_calendar_set_year_prev (GtkCalendar *calendar)
   else
     gtk_calendar_select_day (calendar, calendar->selected_day);
   
-  gtk_calendar_paint (GTK_WIDGET (calendar), NULL);
+  gtk_widget_queue_draw (GTK_WIDGET (calendar));
   gtk_calendar_thaw (calendar);
 }
 
@@ -764,7 +765,7 @@ gtk_calendar_set_year_next (GtkCalendar *calendar)
   else
     gtk_calendar_select_day (calendar, calendar->selected_day);
   
-  gtk_calendar_paint (GTK_WIDGET (calendar), NULL);
+  gtk_widget_queue_draw (GTK_WIDGET (calendar));
   gtk_calendar_thaw (calendar);
 }
 
@@ -856,12 +857,12 @@ gtk_calendar_realize_arrows (GtkWidget *widget)
 			      + private_data->max_month_width);
 	      break;
 	    case ARROW_YEAR_LEFT:
-	      attributes.x = (widget->allocation.width - 4
+	      attributes.x = (widget->allocation.width - 2 * widget->style->xthickness
 			      - (3 + 2*private_data->arrow_width 
 				 + private_data->max_year_width));
 	      break;
 	    case ARROW_YEAR_RIGHT:
-	      attributes.x = (widget->allocation.width - 4 
+	      attributes.x = (widget->allocation.width - 2 * widget->style->xthickness 
 			      - 3 - private_data->arrow_width);
 	      break;
 	    }
@@ -904,10 +905,10 @@ gtk_calendar_realize_header (GtkWidget *widget)
       attributes.colormap = gtk_widget_get_colormap (widget);
       attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
       attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
-      attributes.x = 2;
-      attributes.y = 2;
-      attributes.width = widget->allocation.width - 4;
-      attributes.height = private_data->header_h;
+      attributes.x = widget->style->xthickness;
+      attributes.y = widget->style->ythickness;
+      attributes.width = widget->allocation.width - 2 * attributes.x;
+      attributes.height = private_data->header_h - 2 * attributes.y;
       private_data->header_win = gdk_window_new (widget->window,
 					     &attributes, attributes_mask);
       
@@ -990,7 +991,7 @@ gtk_calendar_realize_week_numbers (GtkWidget *widget)
       attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
       
       attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
-      attributes.x = + (widget->style->xthickness + INNER_BORDER);
+      attributes.x = widget->style->xthickness + INNER_BORDER;
       attributes.y = (private_data->header_h + private_data->day_name_h 
 		      + (widget->style->ythickness + INNER_BORDER));
       attributes.width = private_data->week_width;
@@ -1017,8 +1018,6 @@ gtk_calendar_realize (GtkWidget *widget)
   gint attributes_mask;
   GdkGCValues values;
 
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -1053,11 +1052,7 @@ gtk_calendar_realize (GtkWidget *widget)
 			    | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
 			    | GDK_POINTER_MOTION_MASK | GDK_LEAVE_NOTIFY_MASK);
   
-  if (calendar->display_flags & GTK_CALENDAR_SHOW_WEEK_NUMBERS)
-    attributes.x = private_data->week_width;
-  else
-    attributes.x = 0;
-  attributes.x += (widget->style->xthickness + INNER_BORDER);
+  attributes.x = private_data->week_width + (widget->style->ythickness + INNER_BORDER);
   attributes.y = (private_data->header_h + private_data->day_name_h 
 		  + (widget->style->ythickness + INNER_BORDER));
   attributes.width = (widget->allocation.width - attributes.x 
@@ -1091,8 +1086,6 @@ gtk_calendar_unrealize (GtkWidget *widget)
   GtkCalendarPrivateData *private_data;
   gint i;
   
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -1155,9 +1148,15 @@ gtk_calendar_size_request (GtkWidget	  *widget,
   gint calendar_margin = CALENDAR_MARGIN;
   gint header_width, main_width;
   gint max_header_height = 0;
+  gint focus_width;
+  gint focus_padding;
   
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
+  gtk_widget_style_get (GTK_WIDGET (widget),
+			"focus-line-width", &focus_width,
+			"focus-padding", &focus_padding,
+			NULL);
 
   layout = gtk_widget_create_pango_layout (widget, NULL);
   
@@ -1207,12 +1206,16 @@ gtk_calendar_size_request (GtkWidget	  *widget,
   /* Mainwindow labels width */
   
   private_data->max_day_char_width = 0;
+  private_data->min_day_width = 0;
+  private_data->max_label_char_ascent = 0;
+  private_data->max_label_char_descent = 0;
+
   for (i = 0; i < 9; i++)
     {
       sprintf (buffer, "%d%d", i, i);
       pango_layout_set_text (layout, buffer, -1);	  
       pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
-      private_data->min_day_width = MAX (private_data->max_day_char_width,
+      private_data->min_day_width = MAX (private_data->min_day_width,
 					 logical_rect.width);
 
       private_data->max_day_char_ascent = MAX (private_data->max_label_char_ascent,
@@ -1221,7 +1224,7 @@ gtk_calendar_size_request (GtkWidget	  *widget,
 						PANGO_DESCENT (logical_rect));
     }
   /* We add one to max_day_char_width to be able to make the marked day "bold" */
-  private_data->max_day_char_width = private_data->min_day_width / 2 +1;
+  private_data->max_day_char_width = private_data->min_day_width / 2 + 1;
   
   if (calendar->display_flags & GTK_CALENDAR_SHOW_DAY_NAMES)
     for (i = 0; i < 7; i++)
@@ -1247,13 +1250,13 @@ gtk_calendar_size_request (GtkWidget	  *widget,
 						 logical_rect.width / 2);
       }
   
-  main_width = (7 * (private_data->min_day_width + DAY_XPAD * 2) + (DAY_XSEP * 6) + CALENDAR_MARGIN * 2
+  main_width = (7 * (private_data->min_day_width + (focus_padding + focus_width) * 2) + (DAY_XSEP * 6) + CALENDAR_MARGIN * 2
 		+ (private_data->max_week_char_width
-		   ? private_data->max_week_char_width * 2 + DAY_XPAD * 2 + CALENDAR_XSEP * 2
+		   ? private_data->max_week_char_width * 2 + (focus_padding + focus_width) * 2 + CALENDAR_XSEP * 2
 		   : 0));
   
   
-  requisition->width = MAX (header_width+4, main_width + (widget->style->xthickness + INNER_BORDER) *2);
+  requisition->width = MAX (header_width, main_width + INNER_BORDER * 2) + widget->style->xthickness * 2;
   
   /*
    * Calculate the requisition height for the widget.
@@ -1272,7 +1275,7 @@ gtk_calendar_size_request (GtkWidget	  *widget,
     {
       private_data->day_name_h = (private_data->max_label_char_ascent
 				  + private_data->max_label_char_descent
-				  + 2 * DAY_YPAD + calendar_margin);
+				  + 2 * (focus_padding + focus_width) + calendar_margin);
       calendar_margin = CALENDAR_YSEP;
     } 
   else
@@ -1283,23 +1286,9 @@ gtk_calendar_size_request (GtkWidget	  *widget,
   private_data->main_h = (CALENDAR_MARGIN + calendar_margin
 			  + 6 * (private_data->max_day_char_ascent
 				 + private_data->max_day_char_descent 
-				 + DAY_YPAD * 2)
+				 + 2 * (focus_padding + focus_width))
 			  + DAY_YSEP * 5);
   
-  /* 
-   * If we display weeknumbers we need some extra space 
-   */
-  
-  if (calendar->display_flags & GTK_CALENDAR_SHOW_WEEK_NUMBERS)
-    {
-      private_data->main_h = MAX (private_data->main_h,
-				  (CALENDAR_MARGIN + calendar_margin
-				   + 6 * (private_data->max_day_char_ascent 
-					  + private_data->max_day_char_descent 
-					  + DAY_YPAD * 2)
-				   + DAY_YSEP * 5));
-    }
-  
   height = (private_data->header_h + private_data->day_name_h 
 	    + private_data->main_h);
   
@@ -1314,9 +1303,8 @@ gtk_calendar_size_allocate (GtkWidget	  *widget,
 {
   GtkCalendar *calendar;
   GtkCalendarPrivateData *private_data;
-  
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  g_return_if_fail (allocation != NULL);
+  gint xthickness = widget->style->xthickness;
+  gint ythickness = widget->style->xthickness;
   
   widget->allocation = *allocation;
   
@@ -1326,19 +1314,19 @@ gtk_calendar_size_allocate (GtkWidget	  *widget,
   if (calendar->display_flags & GTK_CALENDAR_SHOW_WEEK_NUMBERS)
     {
       private_data->day_width = (private_data->min_day_width
-			     * ((allocation->width - (widget->style->xthickness + INNER_BORDER) * 2
-				 - (CALENDAR_MARGIN * 2) -  (DAY_XSEP * 7) - CALENDAR_XSEP * 2))
+			     * ((allocation->width - (xthickness + INNER_BORDER) * 2
+				 - (CALENDAR_MARGIN * 2) -  (DAY_XSEP * 6) - CALENDAR_XSEP * 2))
 			     / (7 * private_data->min_day_width + private_data->max_week_char_width * 2));
-      private_data->week_width = ((allocation->width - (widget->style->xthickness + INNER_BORDER) * 2
-			       - (CALENDAR_MARGIN * 2) - (DAY_XSEP * 7) - CALENDAR_XSEP * 2 )
+      private_data->week_width = ((allocation->width - (xthickness + INNER_BORDER) * 2
+			       - (CALENDAR_MARGIN * 2) - (DAY_XSEP * 6) - CALENDAR_XSEP * 2 )
 			      - private_data->day_width * 7 + CALENDAR_MARGIN + CALENDAR_XSEP);
     } 
   else 
     {
       private_data->day_width = (allocation->width
-			     - (widget->style->xthickness + INNER_BORDER) * 2
+			     - (xthickness + INNER_BORDER) * 2
 			     - (CALENDAR_MARGIN * 2)
-			     - (DAY_XSEP * 7))/7;
+			     - (DAY_XSEP * 6))/7;
       private_data->week_width = 0;
     }
   
@@ -1349,8 +1337,8 @@ gtk_calendar_size_allocate (GtkWidget	  *widget,
 			      allocation->width, allocation->height);
       if (private_data->header_win)
 	gdk_window_move_resize (private_data->header_win,
-				2, 2,
-				allocation->width-4, private_data->header_h);
+				xthickness, ythickness,
+				allocation->width - 2 * xthickness, private_data->header_h);
       if (private_data->arrow_win[ARROW_MONTH_LEFT])
 	gdk_window_move_resize (private_data->arrow_win[ARROW_MONTH_LEFT],
 				3, 3,
@@ -1365,7 +1353,7 @@ gtk_calendar_size_allocate (GtkWidget	  *widget,
 				private_data->header_h - 7);
       if (private_data->arrow_win[ARROW_YEAR_LEFT])
 	gdk_window_move_resize (private_data->arrow_win[ARROW_YEAR_LEFT],
-				(allocation->width - 4
+				(allocation->width - 2 * xthickness
 				 - (3 + 2*private_data->arrow_width 
 				    + private_data->max_year_width)),
 				3,
@@ -1373,67 +1361,35 @@ gtk_calendar_size_allocate (GtkWidget	  *widget,
 				private_data->header_h - 7);
       if (private_data->arrow_win[ARROW_YEAR_RIGHT])
 	gdk_window_move_resize (private_data->arrow_win[ARROW_YEAR_RIGHT],
-				(allocation->width - 4 
+				(allocation->width - 2 * xthickness 
 				 - 3 - private_data->arrow_width), 
 				3,
 				private_data->arrow_width,
 				private_data->header_h - 7);
       if (private_data->day_name_win)
 	gdk_window_move_resize (private_data->day_name_win,
-				widget->style->xthickness + INNER_BORDER,
+				xthickness + INNER_BORDER,
 				private_data->header_h + (widget->style->ythickness + INNER_BORDER),
-				allocation->width - (widget->style->xthickness + INNER_BORDER) * 2,
+				allocation->width - (xthickness + INNER_BORDER) * 2,
 				private_data->day_name_h);
       if (private_data->week_win)
 	gdk_window_move_resize (private_data->week_win,
-				(widget->style->xthickness + INNER_BORDER),
+				(xthickness + INNER_BORDER),
 				private_data->header_h + private_data->day_name_h
 				+ (widget->style->ythickness + INNER_BORDER),
 				private_data->week_width,
 				private_data->main_h);
       gdk_window_move_resize (private_data->main_win,
-			      (private_data->week_width ? private_data->week_width + CALENDAR_XSEP :0) 
-			      + (widget->style->xthickness + INNER_BORDER),
+			      private_data->week_width + (xthickness + INNER_BORDER),
 			      private_data->header_h + private_data->day_name_h
 			      + (widget->style->ythickness + INNER_BORDER),
 			      allocation->width 
-			      - (private_data->week_width ? private_data->week_width + CALENDAR_XSEP :0) 
-			      - (widget->style->xthickness + INNER_BORDER) * 2,
+			      - private_data->week_width 
+			      - (xthickness + INNER_BORDER) * 2,
 			      private_data->main_h);
     }
 }
 
-static void
-gtk_calendar_draw_focus (GtkWidget *widget)
-{
-  GtkCalendar *calendar;
-  GtkCalendarPrivateData *private_data;
-  gint width, height;
-  gint x, y;
-  
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-
-  calendar = GTK_CALENDAR (widget);
-  private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
-  
-  if (GTK_WIDGET_DRAWABLE (widget))
-    {
-      x = 0;
-      y = 0;
-      gdk_window_get_size (widget->window, &width, &height);
-      gdk_window_clear (widget->window);
-      
-      gdk_draw_rectangle (widget->window, 
-			  widget->style->base_gc[GTK_WIDGET_STATE (widget)],
-			  FALSE, x + 2, y + 2, width - 5, height - 5);
-      
-      gtk_draw_shadow (widget->style, widget->window,
-		       GTK_STATE_NORMAL, GTK_SHADOW_IN,
-		       x, y, width, height);
-      
-    }
-}
-
 static gboolean
 gtk_calendar_expose (GtkWidget	    *widget,
 		     GdkEventExpose *event)
@@ -1441,10 +1397,6 @@ gtk_calendar_expose (GtkWidget	    *widget,
   GtkCalendar *calendar;
   GtkCalendarPrivateData *private_data;
 
-  g_return_val_if_fail (widget != NULL, FALSE);
-  g_return_val_if_fail (GTK_IS_CALENDAR (widget), FALSE);
-  g_return_val_if_fail (event != NULL, FALSE);
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -1462,40 +1414,16 @@ gtk_calendar_expose (GtkWidget	    *widget,
       if (event->window == private_data->week_win)
 	gtk_calendar_paint_week_numbers (widget);
       if (event->window == widget->window)
-	gtk_calendar_draw_focus (widget);
+	{
+	  gtk_paint_shadow (widget->style, widget->window, GTK_WIDGET_STATE (widget),
+			    GTK_SHADOW_IN, NULL, widget, "calendar",
+			    0, 0, widget->allocation.width, widget->allocation.height);
+	}
     }
   
   return FALSE;
 }
 
-static void
-gtk_calendar_paint (GtkWidget	 *widget,
-		    GdkRectangle *area)
-{
-  GtkCalendar *calendar;
-  GtkCalendarPrivateData *private_data;
-
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  g_return_if_fail (widget->window != NULL);
-  
-  calendar = GTK_CALENDAR (widget);
-  private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
-  
-  if (private_data->main_win != NULL)
-    gtk_calendar_paint_main (widget);
-  
-  if (private_data->header_win != NULL)
-    gtk_calendar_paint_header (widget);
-  
-  if (private_data->day_name_win != NULL)
-    gtk_calendar_paint_day_names (widget);
-  
-  if (private_data->week_win != NULL)
-    gtk_calendar_paint_week_numbers (widget);
-  
-  gtk_calendar_draw_focus (widget);
-}
-
 static void
 gtk_calendar_paint_header (GtkWidget *widget)
 {
@@ -1524,7 +1452,7 @@ gtk_calendar_paint_header (GtkWidget *widget)
   /* Clear window */
   gdk_window_clear (private_data->header_win);
   
-  header_width = widget->allocation.width - 4;
+  header_width = widget->allocation.width - 2 * widget->style->xthickness;
   cal_height = widget->allocation.height;
   
   max_month_width = private_data->max_month_width;
@@ -1591,12 +1519,18 @@ gtk_calendar_paint_day_names (GtkWidget *widget)
   PangoLayout *layout;
   PangoRectangle logical_rect;
   GtkCalendarPrivateData *private_data;
-
+  gint focus_padding;
+  gint focus_width;
+  
   g_return_if_fail (GTK_IS_CALENDAR (widget));
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   gc = calendar->gc;
   
+  gtk_widget_style_get (GTK_WIDGET (widget),
+			"focus-line-width", &focus_width,
+			"focus-padding", &focus_padding,
+			NULL);
   /*
    * Handle freeze/thaw functionality
    */
@@ -1623,7 +1557,7 @@ gtk_calendar_paint_day_names (GtkWidget *widget)
    * Draw rectangles as inverted background for the labels.
    */
   
-  gdk_gc_set_foreground (gc, &widget->style->bg[GTK_STATE_SELECTED]);
+  gdk_gc_set_foreground (gc, &widget->style->base[GTK_STATE_SELECTED]);
   gdk_draw_rectangle (private_data->day_name_win, gc, TRUE,
 		      CALENDAR_MARGIN, CALENDAR_MARGIN,
 		      cal_width-CALENDAR_MARGIN * 2,
@@ -1642,7 +1576,7 @@ gtk_calendar_paint_day_names (GtkWidget *widget)
 
   layout = gtk_widget_create_pango_layout (widget, NULL);
   
-  gdk_gc_set_foreground (gc, &widget->style->fg[GTK_STATE_SELECTED]);
+  gdk_gc_set_foreground (gc, &widget->style->text[GTK_STATE_SELECTED]);
   for (i = 0; i < 7; i++)
     {
       if (gtk_widget_get_direction (GTK_WIDGET (calendar)) == GTK_TEXT_DIR_RTL)
@@ -1657,11 +1591,12 @@ gtk_calendar_paint_day_names (GtkWidget *widget)
       pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
 
       gdk_draw_layout (private_data->day_name_win, gc, 
-		       ((private_data->week_width ? CALENDAR_XSEP : CALENDAR_MARGIN)
-			+ day_wid_sep * i
+		       (CALENDAR_MARGIN +
 			+ private_data->week_width
+			+ (private_data->week_width ? CALENDAR_XSEP : 0)
+			+ day_wid_sep * i
 			+ (day_width - logical_rect.width)/2),
-		       CALENDAR_MARGIN + DAY_YPAD + logical_rect.y,
+		       CALENDAR_MARGIN + focus_width + focus_padding + logical_rect.y,
 		       layout);
     }
   
@@ -1680,6 +1615,8 @@ gtk_calendar_paint_week_numbers (GtkWidget *widget)
   GtkCalendarPrivateData *private_data;
   PangoLayout *layout;
   PangoRectangle logical_rect;
+  gint focus_padding;
+  gint focus_width;
   
   g_return_if_fail (GTK_IS_CALENDAR (widget));
   g_return_if_fail (widget->window != NULL);
@@ -1698,6 +1635,11 @@ gtk_calendar_paint_week_numbers (GtkWidget *widget)
     }
   private_data->dirty_week = 0;
   
+  gtk_widget_style_get (GTK_WIDGET (widget),
+			"focus-line-width", &focus_width,
+			"focus-padding", &focus_padding,
+			NULL);
+  
   /*
    * Clear the window
    */
@@ -1708,18 +1650,18 @@ gtk_calendar_paint_week_numbers (GtkWidget *widget)
    * Draw a rectangle as inverted background for the labels.
    */
   
-  gdk_gc_set_foreground (gc, &widget->style->bg[GTK_STATE_SELECTED]);
+  gdk_gc_set_foreground (gc, &widget->style->base[GTK_STATE_SELECTED]);
   if (private_data->day_name_win)
     gdk_draw_rectangle (private_data->week_win, gc, TRUE,
 			CALENDAR_MARGIN,
 			0,
-			private_data->week_width - CALENDAR_MARGIN - CALENDAR_XSEP,
+			private_data->week_width - CALENDAR_MARGIN,
 			private_data->main_h - CALENDAR_MARGIN);
   else
     gdk_draw_rectangle (private_data->week_win, gc, TRUE,
 			CALENDAR_MARGIN,
 			CALENDAR_MARGIN,
-			private_data->week_width - CALENDAR_MARGIN - CALENDAR_XSEP,
+			private_data->week_width - CALENDAR_MARGIN,
 			private_data->main_h - 2 * CALENDAR_MARGIN);
   
   /*
@@ -1728,7 +1670,7 @@ gtk_calendar_paint_week_numbers (GtkWidget *widget)
   
   layout = gtk_widget_create_pango_layout (widget, NULL);
   
-  gdk_gc_set_foreground (gc, &widget->style->fg[GTK_STATE_SELECTED]);
+  gdk_gc_set_foreground (gc, &widget->style->text[GTK_STATE_SELECTED]);
   day_height = row_height (calendar);
   for (row = 0; row < 6; row++)
     {
@@ -1748,7 +1690,7 @@ gtk_calendar_paint_week_numbers (GtkWidget *widget)
 
       x_loc = (private_data->week_width
 	       - logical_rect.width
-	       - CALENDAR_XSEP - DAY_XPAD);
+	       - CALENDAR_XSEP - focus_padding - focus_width);
 
       gdk_draw_layout (private_data->week_win, gc, x_loc, y_loc, layout);
     }
@@ -1799,6 +1741,8 @@ gtk_calendar_paint_day (GtkWidget *widget,
   gint y_top;
   gint y_loc;
   gint day_xspace;
+  gint focus_width;
+
   GtkCalendarPrivateData *private_data;
   PangoLayout *layout;
   PangoRectangle logical_rect;
@@ -1819,6 +1763,8 @@ gtk_calendar_paint_day (GtkWidget *widget,
       return;
     }
   
+  gtk_widget_style_get (widget, "focus-line-width", &focus_width, NULL);
+
   day_height = row_height (calendar);
   
   day_xspace = private_data->day_width - private_data->max_day_char_width*2;
@@ -1855,7 +1801,7 @@ gtk_calendar_paint_day (GtkWidget *widget,
       */
       if (calendar->selected_day == day)
 	{
-	  gdk_gc_set_foreground (gc, & (GTK_WIDGET (calendar)->style->bg[GTK_STATE_SELECTED]));
+	  gdk_gc_set_foreground (gc, & (GTK_WIDGET (calendar)->style->base[GTK_STATE_SELECTED]));
 	  gdk_draw_rectangle (private_data->main_win, gc, TRUE, x_left, y_top,
 			      private_data->day_width, day_height);
 	}
@@ -1866,7 +1812,7 @@ gtk_calendar_paint_day (GtkWidget *widget,
 	gdk_gc_set_foreground (gc, NORMAL_DAY_COLOR (GTK_WIDGET (calendar)));
   
       if (calendar->selected_day == day)
-	gdk_gc_set_foreground (gc, & (GTK_WIDGET (calendar)->style->fg[GTK_STATE_SELECTED]));
+	gdk_gc_set_foreground (gc, & (GTK_WIDGET (calendar)->style->text[GTK_STATE_SELECTED]));
       else
 	gdk_gc_set_foreground (gc, & (GTK_WIDGET (calendar)->style->fg[GTK_WIDGET_STATE (calendar)]));
     }
@@ -1889,9 +1835,14 @@ gtk_calendar_paint_day (GtkWidget *widget,
   if (GTK_WIDGET_HAS_FOCUS (calendar) 
       && calendar->focus_row == row && calendar->focus_col == col)
     {
-      gdk_draw_rectangle (private_data->main_win, calendar->xor_gc, 
-			  FALSE, x_left, y_top,
-			  private_data->day_width-1, day_height-1);
+      gtk_paint_focus (widget->style, 
+		       private_data->main_win,
+		       (calendar->selected_day == day) 
+		          ? GTK_STATE_SELECTED : GTK_STATE_NORMAL, 
+		       NULL, widget, "calendar-day",
+		       x_left, y_top, 
+		       private_data->day_width, 
+		       day_height);
     }
 
   g_object_unref (G_OBJECT (layout));
@@ -2143,7 +2094,6 @@ gtk_calendar_select_month (GtkCalendar *calendar,
 			   guint	month,
 			   guint	year)
 {
-  g_return_val_if_fail (calendar != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_CALENDAR (calendar), FALSE);
   g_return_val_if_fail (month <= 11, FALSE);
   
@@ -2152,9 +2102,7 @@ gtk_calendar_select_month (GtkCalendar *calendar,
   
   gtk_calendar_compute_days (calendar);
   
-  if (GTK_WIDGET_DRAWABLE (GTK_WIDGET (calendar)))
-    gtk_calendar_paint (GTK_WIDGET (calendar), NULL);
-  
+  gtk_widget_queue_draw (GTK_WIDGET (calendar));
   gtk_signal_emit (GTK_OBJECT (calendar),
 		   gtk_calendar_signals[MONTH_CHANGED_SIGNAL]);
   return TRUE;
@@ -2182,7 +2130,7 @@ gtk_calendar_select_day (GtkCalendar *calendar,
   
   calendar->selected_day = day;
   
-  /* Deselect the new day */
+  /* Select the new day */
   if (day != 0)
     {
       if (GTK_WIDGET_DRAWABLE (GTK_WIDGET (calendar)))
@@ -2217,7 +2165,6 @@ gboolean
 gtk_calendar_mark_day (GtkCalendar *calendar,
 		       guint	    day)
 {
-  g_return_val_if_fail (calendar != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_CALENDAR (calendar), FALSE);
   
   if (day >= 1 && day <= 31 && calendar->marked_date[day-1] == FALSE)
@@ -2237,7 +2184,6 @@ gboolean
 gtk_calendar_unmark_day (GtkCalendar *calendar,
 			 guint	      day)
 {
-  g_return_val_if_fail (calendar != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_CALENDAR (calendar), FALSE);
   
   if (day >= 1 && day <= 31 && calendar->marked_date[day-1] == TRUE)
@@ -2281,10 +2227,6 @@ gtk_calendar_button_press (GtkWidget	  *widget,
   gint x, y;
   void (* action_func) (GtkCalendar *);
   
-  g_return_val_if_fail (widget != NULL, FALSE);
-  g_return_val_if_fail (GTK_IS_CALENDAR (widget), FALSE);
-  g_return_val_if_fail (event != NULL, FALSE);
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -2361,9 +2303,6 @@ gtk_calendar_enter_notify (GtkWidget	    *widget,
   GtkCalendar *calendar;
   GtkCalendarPrivateData *private_data;
   
-  g_return_val_if_fail (widget != NULL, FALSE);
-  g_return_val_if_fail (event != NULL, FALSE);
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -2403,9 +2342,6 @@ gtk_calendar_leave_notify (GtkWidget	    *widget,
   gint row;
   gint col;
   
-  g_return_val_if_fail (widget != NULL, FALSE);
-  g_return_val_if_fail (event != NULL, FALSE);
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
   
@@ -2457,8 +2393,6 @@ gtk_calendar_paint_arrow (GtkWidget *widget,
   gint state;
   gint width, height;
   
-  g_return_if_fail (widget != NULL);
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
 
@@ -2535,8 +2469,6 @@ gtk_calendar_set_background (GtkWidget *widget)
   GtkCalendarPrivateData *private_data;
   gint i;
   
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  
   calendar = GTK_CALENDAR (widget);
   private_data = GTK_CALENDAR_PRIVATE_DATA (widget);
 
@@ -2570,8 +2502,6 @@ static void
 gtk_calendar_style_set (GtkWidget *widget,
 			GtkStyle  *previous_style)
 {
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  
   if (previous_style && GTK_WIDGET_REALIZED (widget))
     gtk_calendar_set_background(widget);
 }
@@ -2580,19 +2510,19 @@ static void
 gtk_calendar_state_changed (GtkWidget	   *widget,
 			    GtkStateType    previous_state)
 {
-  g_return_if_fail (GTK_IS_CALENDAR (widget));
-  
   gtk_calendar_set_background (widget);
 }
 
 static void
-gtk_calendar_destroy (GtkObject *object)
+gtk_calendar_finalize (GObject *object)
 {
   GtkCalendarPrivateData *private_data;
   private_data = GTK_CALENDAR_PRIVATE_DATA (object);
+  
   g_free (private_data);
-  if (GTK_OBJECT_CLASS (parent_class)->destroy)
-    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
+  
+  if (G_OBJECT_CLASS (parent_class)->finalize)
+    (* G_OBJECT_CLASS (parent_class)->finalize) (object);
 }
 
 static gboolean
@@ -2604,10 +2534,6 @@ gtk_calendar_key_press (GtkWidget   *widget,
   gint old_focus_row;
   gint old_focus_col;
   gint row, col, day;
-
-  g_return_val_if_fail (widget != NULL, FALSE);
-  g_return_val_if_fail (GTK_IS_CALENDAR (widget), FALSE);
-  g_return_val_if_fail (event != NULL, FALSE);
   
   calendar = GTK_CALENDAR (widget);
   return_val = FALSE;
