@@ -24,6 +24,10 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+/* Do NOT, I repeat, NOT, copy any of the code in this file.
+ * The code here relies on all sorts of internal details of GTK+
+ */
+
 #include <string.h>
 
 #include "gtkarrow.h"
@@ -86,7 +90,7 @@ static void         gtk_combo_update_list        (GtkEntry         *entry,
 static gint         gtk_combo_button_press       (GtkWidget        *widget,
 						  GdkEvent         *event,
 						  GtkCombo         *combo);
-static gint         gtk_combo_button_release     (GtkWidget        *widget,
+static void         gtk_combo_button_event_after (GtkWidget        *widget,
 						  GdkEvent         *event,
 						  GtkCombo         *combo);
 static gint         gtk_combo_list_enter         (GtkWidget        *widget,
@@ -195,7 +199,6 @@ gtk_combo_entry_key_press (GtkEntry * entry, GdkEventKey * event, GtkCombo * com
 
   /* completion */
   if ((event->keyval == GDK_Tab ||
-       event->keyval == GDK_ISO_Left_Tab ||
        event->keyval == GDK_KP_Tab) &&
       (event->state & GDK_MOD1_MASK)) 
     {
@@ -305,29 +308,41 @@ static GtkListItem *
 gtk_combo_find (GtkCombo * combo)
 {
   const gchar *text;
+  GtkListItem *found = NULL;
   gchar *ltext;
+  gchar *compare_text;
   GList *clist;
-  int (*string_compare) (const char *, const char *);
-
-  if (combo->case_sensitive)
-    string_compare = strcmp;
-  else
-    string_compare = g_strcasecmp;
 
   text = gtk_entry_get_text (GTK_ENTRY (combo->entry));
+  if (combo->case_sensitive)
+    compare_text = (gchar *)text;
+  else
+    compare_text = g_utf8_casefold (text, -1);
+  
   clist = GTK_LIST (combo->list)->children;
 
-  while (clist && clist->data)
+  while (!found && clist && clist->data)
     {
       ltext = gtk_combo_func (GTK_LIST_ITEM (clist->data));
       if (!ltext)
 	continue;
-      if (!(*string_compare) (ltext, text))
-	return (GtkListItem *) clist->data;
+
+      if (!combo->case_sensitive)
+	ltext = g_utf8_casefold (ltext, -1);
+
+      if (strcmp (ltext, compare_text) == 0)
+	found = clist->data;
+
+      if (!combo->case_sensitive)
+	g_free (ltext);
+
       clist = clist->next;
     }
 
-  return NULL;
+  if (!combo->case_sensitive)
+    g_free (compare_text);
+
+  return found;
 }
 
 static gchar *
@@ -486,15 +501,33 @@ gtk_combo_popup_list (GtkCombo * combo)
       gtk_widget_hide (GTK_SCROLLED_WINDOW (combo->popup)->vscrollbar);
     }
 
-  gtk_widget_set_uposition (combo->popwin, x, y);
+  gtk_window_move (GTK_WINDOW (combo->popwin), x, y);
   gtk_widget_set_usize (combo->popwin, width, height);
-  gtk_widget_realize (combo->popwin);
-  gdk_window_resize (combo->popwin->window, width, height);
   gtk_widget_show (combo->popwin);
 
   gtk_widget_grab_focus (combo->popwin);
 }
 
+static void
+gtk_combo_popdown_list (GtkCombo *combo)
+{
+  combo->current_button = 0;
+      
+  if (GTK_BUTTON (combo->button)->in_button)
+    {
+      GTK_BUTTON (combo->button)->in_button = FALSE;
+      gtk_button_released (GTK_BUTTON (combo->button));
+    }
+
+  if (GTK_WIDGET_HAS_GRAB (combo->popwin))
+    {
+      gtk_grab_remove (combo->popwin);
+      gdk_pointer_ungrab (GDK_CURRENT_TIME);
+    }
+  
+  gtk_widget_hide (combo->popwin);
+}
+
 static void        
 gtk_combo_activate (GtkWidget        *widget,
 		    GtkCombo         *combo)
@@ -546,10 +579,12 @@ gtk_combo_popup_button_leave (GtkWidget        *button,
 			      GdkEventCrossing *event,
 			      GtkCombo         *combo)
 {
+  /* The idea here is that we want to keep the button down if the
+   * popup is popped up.
+   */
   return combo->current_button != 0;
 }
 
-
 static void
 gtk_combo_update_entry (GtkList * list, GtkCombo * combo)
 {
@@ -609,52 +644,27 @@ gtk_combo_button_press (GtkWidget * widget, GdkEvent * event, GtkCombo * combo)
 	}
     }
 
-  gtk_widget_hide (combo->popwin);
-  gtk_grab_remove (combo->popwin);
-  gdk_pointer_ungrab (event->button.time);
+  gtk_combo_popdown_list (combo);
 
   return TRUE;
 }
 
-static gint
-gtk_combo_button_release (GtkWidget *widget,
-			  GdkEvent  *event,
-			  GtkCombo  *combo)
+static void
+gtk_combo_button_event_after (GtkWidget *widget,
+			      GdkEvent  *event,
+			      GtkCombo  *combo)
 {
   GtkWidget *child;
 
-  /* Horrible hack to get connect-after effect without regard to the return value of the default
-   * handler.
-   */
-  gtk_signal_handler_block_by_func (GTK_OBJECT (widget),
-				    GTK_SIGNAL_FUNC (gtk_combo_button_release),
-				    combo);
-  gtk_widget_event (widget, event);
-  gtk_signal_handler_unblock_by_func (GTK_OBJECT (widget),
-				      GTK_SIGNAL_FUNC (gtk_combo_button_release),
-				      combo);
+  if (event->type != GDK_BUTTON_RELEASE)
+    return;
   
   if ((combo->current_button != 0) && (event->button.button == 1))
     {
       /* This was the initial button press */
 
-      GdkEventCrossing tmp_event;
-
       combo->current_button = 0;
 
-      if (widget != combo->button)
-	gtk_widget_event (combo->button, event);
-
-      /* Un-pre-highlight */
-      
-      tmp_event.type = GDK_LEAVE_NOTIFY;
-      tmp_event.window = combo->button->window;
-      tmp_event.send_event = TRUE;
-      tmp_event.subwindow = NULL;
-      tmp_event.detail = GDK_NOTIFY_ANCESTOR;
-      
-      gtk_widget_event (combo->button, (GdkEvent *)&tmp_event);
-
       /* Check to see if we released inside the button */
       child = gtk_get_event_widget ((GdkEvent*) event);
 
@@ -669,23 +679,11 @@ gtk_combo_button_release (GtkWidget *widget,
 			    GDK_BUTTON_RELEASE_MASK |
 			    GDK_POINTER_MOTION_MASK, 
 			    NULL, NULL, GDK_CURRENT_TIME);
-	  return TRUE;
+	  return;
 	}
     }
-  else
-    {
-      /* The user has clicked inside the popwin and released */
 
-      if (GTK_WIDGET_HAS_GRAB (combo->popwin))
-	{
-	  gtk_grab_remove (combo->popwin);
-	  gdk_pointer_ungrab (event->button.time);
-	}
-    }
-  
-  gtk_widget_hide (combo->popwin);
-
-  return TRUE;
+  gtk_combo_popdown_list (combo);
 }
 
 static gint         
@@ -735,21 +733,11 @@ gtk_combo_list_key_press (GtkWidget * widget, GdkEventKey * event, GtkCombo * co
 {
   if (event->keyval == GDK_Escape)
     {
-      if (GTK_WIDGET_HAS_GRAB (combo->popwin))
-	{
-	  gtk_grab_remove (combo->popwin);
-	  gdk_pointer_ungrab (GDK_CURRENT_TIME);
-	}
-      else if (GTK_WIDGET_HAS_GRAB (combo->list))
+      if (GTK_WIDGET_HAS_GRAB (combo->list))
 	gtk_list_end_drag_selection (GTK_LIST (combo->list));
-      gtk_widget_hide (combo->popwin);
-      if (GTK_WIDGET_HAS_GRAB (combo->button))
-	{
-	  combo->current_button = 0;
-	  GTK_BUTTON (combo->button)->in_button = FALSE;
-	  gtk_button_released (GTK_BUTTON (combo->button));
-	  gtk_grab_remove (combo->button);
-	}
+
+      gtk_combo_popdown_list (combo);
+      
       return TRUE;
     }
   return FALSE;
@@ -789,16 +777,12 @@ gtk_combo_init (GtkCombo * combo)
 		      (GtkSignalFunc) gtk_combo_activate, combo);
   gtk_signal_connect (GTK_OBJECT (combo->button), "button_press_event",
 		      (GtkSignalFunc) gtk_combo_popup_button_press, combo);
-  /*gtk_signal_connect_after (GTK_OBJECT (combo->button), "button_release_event",
-    (GtkSignalFunc) gtk_combo_button_release, combo);*/
   gtk_signal_connect (GTK_OBJECT (combo->button), "leave_notify_event",
 		      (GtkSignalFunc) gtk_combo_popup_button_leave, combo);
-  /*gtk_signal_connect (GTK_OBJECT (combo->button), "clicked",
-     (GtkSignalFunc)prelight_bug, combo); */
 
   combo->popwin = gtk_window_new (GTK_WINDOW_POPUP);
   gtk_widget_ref (combo->popwin);
-  gtk_window_set_policy (GTK_WINDOW (combo->popwin), 1, 1, 0);
+  gtk_window_set_resizable (GTK_WINDOW (combo->popwin), FALSE);
 
   gtk_signal_connect (GTK_OBJECT (combo->popwin), "key_press_event",
 		      GTK_SIGNAL_FUNC (gtk_combo_window_key_press), combo);
@@ -848,8 +832,8 @@ gtk_combo_init (GtkCombo * combo)
   gtk_signal_connect (GTK_OBJECT (combo->popwin), "button_press_event",
 		      GTK_SIGNAL_FUNC (gtk_combo_button_press), combo);
 
-  gtk_signal_connect (GTK_OBJECT (combo->list), "button_release_event",
-		      GTK_SIGNAL_FUNC (gtk_combo_button_release), combo);
+  gtk_signal_connect (GTK_OBJECT (combo->list), "event_after",
+		      (GtkSignalFunc) gtk_combo_button_event_after, combo);
   /* We connect here on the button, because we'll have a grab on it
    * when the event occurs. But we are actually interested in enters
    * for the combo->list.
@@ -891,12 +875,20 @@ void
 gtk_combo_set_value_in_list (GtkCombo * combo, gboolean val, gboolean ok_if_empty)
 {
   g_return_if_fail (GTK_IS_COMBO (combo));
+  val = val != FALSE;
+  ok_if_empty = ok_if_empty != FALSE;
 
-  combo->value_in_list = val;
-  combo->ok_if_empty = ok_if_empty;
   g_object_freeze_notify (G_OBJECT (combo));
+  if (combo->value_in_list != val)
+    {
+       combo->value_in_list = val;
   g_object_notify (G_OBJECT (combo), "value_in_list");
+    }
+  if (combo->ok_if_empty != ok_if_empty)
+    {
+       combo->ok_if_empty = ok_if_empty;
   g_object_notify (G_OBJECT (combo), "allow_empty");
+    }
   g_object_thaw_notify (G_OBJECT (combo));
 }
 
@@ -904,31 +896,47 @@ void
 gtk_combo_set_case_sensitive (GtkCombo * combo, gboolean val)
 {
   g_return_if_fail (GTK_IS_COMBO (combo));
+  val = val != FALSE;
 
+  if (combo->case_sensitive != val) 
+    {
   combo->case_sensitive = val;
   g_object_notify (G_OBJECT (combo), "case_sensitive");
+    }
 }
 
 void
 gtk_combo_set_use_arrows (GtkCombo * combo, gboolean val)
 {
   g_return_if_fail (GTK_IS_COMBO (combo));
+  val = val != FALSE;
 
+  if (combo->use_arrows != val) 
+    {
   combo->use_arrows = val;
   g_object_notify (G_OBJECT (combo), "enable_arrow_keys");
+    }
 }
 
 void
 gtk_combo_set_use_arrows_always (GtkCombo * combo, gboolean val)
 {
   g_return_if_fail (GTK_IS_COMBO (combo));
+  val = val != FALSE;
 
+  if (combo->use_arrows_always != val) 
+    {
+       g_object_freeze_notify (G_OBJECT (combo));
   combo->use_arrows_always = val;
+       g_object_notify (G_OBJECT (combo), "enable_arrows_always");
+
+       if (combo->use_arrows != TRUE) 
+         {
   combo->use_arrows = TRUE;
-  g_object_freeze_notify (G_OBJECT (combo));
   g_object_notify (G_OBJECT (combo), "enable_arrow_keys");
-  g_object_notify (G_OBJECT (combo), "enable_arrows_always");
+         }
   g_object_thaw_notify (G_OBJECT (combo));
+    }
 }
 
 void
@@ -940,6 +948,8 @@ gtk_combo_set_popdown_strings (GtkCombo * combo, GList * strings)
   g_return_if_fail (GTK_IS_COMBO (combo));
   g_return_if_fail (strings != NULL);
 
+  gtk_combo_popdown_list (combo);
+
   gtk_list_clear_items (GTK_LIST (combo->list), 0, -1);
   list = strings;
   while (list)
