@@ -742,17 +742,38 @@ gtk_menu_popup (GtkMenu		    *menu,
   menu->position_func_data = data;
   menu_shell->activate_time = activate_time;
 
-  gtk_menu_position (menu);
-
-  /* We need to show the menu _here_ because code expects to be
-   * able to tell if the menu is onscreen by looking at the
-   * GTK_WIDGET_VISIBLE (menu)
+  /* We need to show the menu here rather in the init function because
+   * code expects to be able to tell if the menu is onscreen by
+   * looking at the GTK_WIDGET_VISIBLE (menu)
    */
   gtk_widget_show (GTK_WIDGET (menu));
-  gtk_widget_show (menu->toplevel);
+
+  /* Compute the size of the toplevel and realize it so we
+   * can position and scroll correctly.
+   */
+  {
+    GtkRequisition tmp_request;
+    GtkAllocation tmp_allocation = { 0, };
+
+    gtk_widget_size_request (menu->toplevel, &tmp_request);
+    
+    tmp_allocation.width = tmp_request.width;
+    tmp_allocation.height = tmp_request.height;
+
+    gtk_widget_size_allocate (menu->toplevel, &tmp_allocation);
+    
+    gtk_widget_realize (GTK_WIDGET (menu));
+  }
+
+  gtk_menu_position (menu);
 
   gtk_menu_scroll_to (menu, menu->scroll_offset);
 
+  /* Once everything is set up correctly, map the toplevel window on
+     the screen.
+   */
+  gtk_widget_show (menu->toplevel);
+
   if (xgrab_shell == widget)
     popup_grab_on_window (widget->window, activate_time); /* Should always succeed */
 
@@ -1297,8 +1318,9 @@ gtk_menu_realize (GtkWidget *widget)
   gtk_style_set_background (widget->style, menu->view_window, GTK_STATE_NORMAL);
   gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);
 
-  gtk_menu_scroll_item_visible (GTK_MENU_SHELL (widget),
-				GTK_MENU_SHELL (widget)->active_menu_item);
+  if (GTK_MENU_SHELL (widget)->active_menu_item)
+    gtk_menu_scroll_item_visible (GTK_MENU_SHELL (widget),
+				  GTK_MENU_SHELL (widget)->active_menu_item);
 
   gdk_window_show (menu->bin_window);
   gdk_window_show (menu->view_window);
@@ -1711,6 +1733,7 @@ gtk_menu_key_press (GtkWidget	*widget,
   gboolean can_change_accels;
   gchar *accel = NULL;
   guint accel_key, accel_mods;
+  GdkModifierType consumed_modifiers;
   
   g_return_val_if_fail (GTK_IS_MENU (widget), FALSE);
   g_return_val_if_fail (event != NULL, FALSE);
@@ -1769,10 +1792,21 @@ gtk_menu_key_press (GtkWidget	*widget,
   g_object_get (G_OBJECT (gtk_settings_get_default ()),
 		"gtk-can-change-accels", &can_change_accels,
 		NULL);
-  
-  accel_key = event->keyval;
-  accel_mods = event->state & gtk_accelerator_get_default_mod_mask ();
 
+  /* Figure out what modifiers went into determining the key symbol */
+  gdk_keymap_translate_keyboard_state (gdk_keymap_get_default (),
+				       event->hardware_keycode, event->state, event->group,
+				       NULL, NULL, NULL, &consumed_modifiers);
+
+  accel_key = gdk_keyval_to_lower (event->keyval);
+  accel_mods = event->state & gtk_accelerator_get_default_mod_mask () & ~consumed_modifiers;
+
+  /* If lowercasing affects the keysym, then we need to include SHIFT in the modifiers,
+   * We re-upper case when we match against the keyval, but display and save in caseless form.
+   */
+  if (accel_key != event->keyval)
+    accel_mods |= GDK_SHIFT_MASK;
+  
   /* Modify the accelerators */
   if (can_change_accels &&
       menu_shell->active_menu_item &&
@@ -1838,7 +1872,6 @@ gtk_menu_motion_notify  (GtkWidget	   *widget,
 
   gboolean need_enter;
 
-
   if (GTK_IS_MENU (widget))
     gtk_menu_handle_scrolling (GTK_MENU (widget), TRUE);
   
@@ -2220,7 +2253,7 @@ gtk_menu_set_submenu_navigation_region (GtkMenu          *menu,
       /* Exiting the top or bottom? */ 
       if (event->y < 0)
         { /* top */
-	  point[0].y = event->y_root + SUBMENU_NAV_REGION_PADDING;
+	  point[0].y = event->y_root + 1;
 	  point[1].y = submenu_top;
 
 	  if (point[0].y <= point[1].y)
@@ -2228,7 +2261,7 @@ gtk_menu_set_submenu_navigation_region (GtkMenu          *menu,
 	}
       else
         { /* bottom */
-	  point[0].y = event->y_root - SUBMENU_NAV_REGION_PADDING; 
+	  point[0].y = event->y_root;
 	  point[1].y = submenu_bottom;
 
 	  if (point[0].y >= point[1].y)
