@@ -56,6 +56,12 @@ struct _GtkScrollbar
 struct _GtkScrollbarClass
 {
   GtkRangeClass parent_class;
+
+  /* Padding for future expansion */
+  void (*_gtk_reserved1) (void);
+  void (*_gtk_reserved2) (void);
+  void (*_gtk_reserved3) (void);
+  void (*_gtk_reserved4) (void);
 };
 
 
