@@ -92,10 +92,11 @@ static void gtk_tree_view_column_update_button                 (GtkTreeViewColum
 static gint gtk_tree_view_column_button_event                  (GtkWidget               *widget,
 								GdkEvent                *event,
 								gpointer                 data);
-static void gtk_tree_view_column_button_realize                (GtkWidget               *widget,
-								gpointer                 data);
 static void gtk_tree_view_column_button_clicked                (GtkWidget               *widget,
 								gpointer                 data);
+static gboolean gtk_tree_view_column_mnemonic_activate         (GtkWidget *widget,
+					                        gboolean   group_cycling,
+								gpointer   data);
 
 /* Property handlers */
 static void gtk_tree_view_model_sort_column_changed            (GtkTreeSortable         *sortable,
@@ -328,8 +329,14 @@ gtk_tree_view_column_finalize (GObject *object)
   for (list = tree_column->cell_list; list; list = list->next)
     {
       GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
-      if (info->func_data && info->destroy)
-	(info->destroy) (info->func_data);
+
+      if (info->destroy)
+	{
+	  GtkDestroyNotify d = info->destroy;
+
+	  info->destroy = NULL;
+	  d (info->func_data);
+	}
       gtk_tree_view_column_clear_attributes (tree_column, info->cell);
       g_object_unref (G_OBJECT (info->cell));
       g_free (info);
@@ -530,24 +537,21 @@ gtk_tree_view_column_create_button (GtkTreeViewColumn *tree_column)
 
   gtk_widget_push_composite_child ();
   tree_column->button = gtk_button_new ();
+  gtk_widget_add_events (tree_column->button, GDK_POINTER_MOTION_MASK);
   gtk_widget_pop_composite_child ();
 
   /* make sure we own a reference to it as well. */
   if (tree_view->priv->header_window)
     gtk_widget_set_parent_window (tree_column->button, tree_view->priv->header_window);
   gtk_widget_set_parent (tree_column->button, GTK_WIDGET (tree_view));
-  
-  gtk_signal_connect (GTK_OBJECT (tree_column->button), "realize",
-		      (GtkSignalFunc) gtk_tree_view_column_button_realize,
-		      NULL);
 
-  gtk_signal_connect (GTK_OBJECT (tree_column->button), "event",
-		      (GtkSignalFunc) gtk_tree_view_column_button_event,
-		      (gpointer) tree_column);
-  
-  gtk_signal_connect (GTK_OBJECT (tree_column->button), "clicked",
-		      (GtkSignalFunc) gtk_tree_view_column_button_clicked,
-		      (gpointer) tree_column);
+  g_signal_connect (G_OBJECT (tree_column->button), "event",
+		    G_CALLBACK (gtk_tree_view_column_button_event),
+		    (gpointer) tree_column);
+
+  g_signal_connect (G_OBJECT (tree_column->button), "clicked",
+		    (GtkSignalFunc) gtk_tree_view_column_button_clicked,
+		    (gpointer) tree_column);
 
   tree_column->alignment = gtk_alignment_new (tree_column->xalign, 0.5, 0.0, 0.0);
 
@@ -562,6 +566,10 @@ gtk_tree_view_column_create_button (GtkTreeViewColumn *tree_column)
       gtk_widget_show (child);
     }
 
+  g_signal_connect (G_OBJECT (child), "mnemonic_activate",
+		    G_CALLBACK (gtk_tree_view_column_mnemonic_activate),
+		    (gpointer) tree_column);
+
   if (tree_column->xalign <= 0.5)
     gtk_box_pack_end (GTK_BOX (hbox), tree_column->arrow, FALSE, FALSE, 0);
   else
@@ -627,11 +635,11 @@ gtk_tree_view_column_update_button (GtkTreeViewColumn *tree_column)
       g_return_if_fail (GTK_IS_LABEL (current_child));
 
       if (tree_column->title)
-	gtk_label_set_text (GTK_LABEL (current_child),
-			    tree_column->title);
+	gtk_label_set_text_with_mnemonic (GTK_LABEL (current_child),
+					  tree_column->title);
       else
-	gtk_label_set_text (GTK_LABEL (current_child),
-			    "");
+	gtk_label_set_text_with_mnemonic (GTK_LABEL (current_child),
+					  "");
     }
 
   switch (tree_column->sort_order)
@@ -723,6 +731,12 @@ gtk_tree_view_column_update_button (GtkTreeViewColumn *tree_column)
 	    }
 	}
     }
+  /* Queue a resize on the assumption that we always want to catch all changes
+   * and columns don't change all that often.
+   */
+  if (GTK_WIDGET_REALIZED (tree_column->tree_view))
+     gtk_widget_queue_resize (tree_column->tree_view);
+
 }
 
 /* Button signal handlers
@@ -761,6 +775,10 @@ gtk_tree_view_column_button_event (GtkWidget *widget,
 				 (gint) ((GdkEventMotion *)event)->y)))
     {
       column->maybe_reordered = FALSE;
+      /* this is to change our drag_x to be relative to
+       * tree_view->priv->bin_window, instead of our window.
+       */
+      column->drag_x -= column->button->allocation.x;
       _gtk_tree_view_column_start_drag (GTK_TREE_VIEW (column->tree_view), column);
       return TRUE;
     }
@@ -783,11 +801,6 @@ gtk_tree_view_column_button_event (GtkWidget *widget,
   return FALSE;
 }
 
-static void
-gtk_tree_view_column_button_realize (GtkWidget *widget, gpointer data)
-{
-  gdk_window_set_events (widget->window, gdk_window_get_events (widget->window) | GDK_POINTER_MOTION_MASK);
-}
 
 static void
 gtk_tree_view_column_button_clicked (GtkWidget *widget, gpointer data)
@@ -795,6 +808,30 @@ gtk_tree_view_column_button_clicked (GtkWidget *widget, gpointer data)
   g_signal_emit_by_name (G_OBJECT (data), "clicked");
 }
 
+static gboolean
+gtk_tree_view_column_mnemonic_activate (GtkWidget *widget,
+					gboolean   group_cycling,
+					gpointer   data)
+{
+  GtkTreeViewColumn *column = (GtkTreeViewColumn *)data;
+
+  g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (column), FALSE);
+
+  if (column->clickable)
+    gtk_button_clicked (GTK_BUTTON (column->button));
+  else if (GTK_WIDGET_CAN_FOCUS (column->button))
+    gtk_widget_grab_focus (column->button);
+  else
+    {
+      GTK_TREE_VIEW (column->tree_view)->priv->focus_column = column;
+      GTK_TREE_VIEW_SET_FLAG (GTK_TREE_VIEW (column->tree_view),
+			      GTK_TREE_VIEW_DRAW_KEYFOCUS);
+      gtk_widget_grab_focus (column->tree_view);
+    }
+
+  return TRUE;
+}
+
 static void
 gtk_tree_view_model_sort_column_changed (GtkTreeSortable   *sortable,
 					 GtkTreeViewColumn *column)
@@ -1078,7 +1115,9 @@ gtk_tree_view_column_get_cell_info (GtkTreeViewColumn *tree_column,
  * @tree_column: A #GtkTreeViewColumn.
  * @cell: The #GtkCellRenderer. 
  * @expand: %TRUE if @cell is to be given extra space allocated to box.
- * 
+ *
+ * Packs the @cell into the beginning column.  If @expand is %TRUE, then the
+ * @cell is allocated a share of all available space that the @tree_column has.
  **/
 void
 gtk_tree_view_column_pack_start (GtkTreeViewColumn *tree_column,
@@ -1104,6 +1143,15 @@ gtk_tree_view_column_pack_start (GtkTreeViewColumn *tree_column,
   tree_column->cell_list = g_list_append (tree_column->cell_list, cell_info);
 }
 
+/**
+ * gtk_tree_view_column_pack_end:
+ * @tree_column: A #GtkTreeViewColumn.
+ * @cell: The #GtkCellRenderer. 
+ * @expand: %TRUE if @cell is to be given extra space allocated to box.
+ *
+ * Packs the @cell into the column.  If @expand is %TRUE, then the @cell is
+ * allocated a share of all available space that the @tree_column has.
+ **/
 void
 gtk_tree_view_column_pack_end (GtkTreeViewColumn  *tree_column,
 			       GtkCellRenderer    *cell,
@@ -1210,7 +1258,7 @@ gtk_tree_view_column_add_attribute (GtkTreeViewColumn *tree_column,
   info->attributes = g_slist_prepend (info->attributes, g_strdup (attribute));
 
   if (tree_column->tree_view)
-    gtk_tree_view_column_cell_set_dirty (tree_column);
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
 
 }
 
@@ -1291,20 +1339,20 @@ gtk_tree_view_column_set_cell_data_func (GtkTreeViewColumn   *tree_column,
 
   g_return_if_fail (info != NULL);
 
-  if (func == info->func &&
-      func_data == info->func_data &&
-      destroy == info->destroy)
-    return;
+  if (info->destroy)
+    {
+      GtkDestroyNotify d = info->destroy;
 
-  if (info->func_data && info->destroy)
-    (info->destroy) (info->func_data);
+      info->destroy = NULL;
+      d (info->func_data);
+    }
 
   info->func = func;
   info->func_data = func_data;
   info->destroy = destroy;
 
   if (tree_column->tree_view)
-    gtk_tree_view_column_cell_set_dirty (tree_column);
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
 }
 
 
@@ -1338,7 +1386,7 @@ gtk_tree_view_column_clear_attributes (GtkTreeViewColumn *tree_column,
   info->attributes = NULL;
 
   if (tree_column->tree_view)
-    gtk_tree_view_column_cell_set_dirty (tree_column);
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
 }
 
 
@@ -1362,7 +1410,7 @@ gtk_tree_view_column_set_spacing (GtkTreeViewColumn *tree_column,
 
   tree_column->spacing = spacing;
   if (tree_column->tree_view)
-    gtk_tree_view_column_cell_set_dirty (tree_column);
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
 }
 
 /**
@@ -1403,6 +1451,9 @@ gtk_tree_view_column_set_visible (GtkTreeViewColumn *tree_column,
 
   tree_column->visible = visible;
 
+  if (tree_column->visible)
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
+
   gtk_tree_view_column_update_button (tree_column);
   g_object_notify (G_OBJECT (tree_column), "visible");
 }
@@ -1424,6 +1475,16 @@ gtk_tree_view_column_get_visible (GtkTreeViewColumn *tree_column)
   return tree_column->visible;
 }
 
+/**
+ * gtk_tree_view_column_set_resizable:
+ * @tree_column: A #GtkTreeViewColumn
+ * @resizable: %TRUE, if the column can be resized
+ * 
+ * If @resizable is %TRUE, then the user can explicitly resize the column by
+ * grabbing the outer edge of the column button.  If resizable is TRUE and
+ * sizing mode of the column is #GTK_TREE_VIEW_COLUMN_AUTOSIZE, then the sizing
+ * mode is changed to #GTK_TREE_VIEW_COLUMN_GROW_ONLY.
+ **/
 void
 gtk_tree_view_column_set_resizable (GtkTreeViewColumn *tree_column,
 				    gboolean           resizable)
@@ -1445,6 +1506,14 @@ gtk_tree_view_column_set_resizable (GtkTreeViewColumn *tree_column,
   g_object_notify (G_OBJECT (tree_column), "resizable");
 }
 
+/**
+ * gtk_tree_view_column_get_resizable:
+ * @tree_column: A #GtkTreeViewColumn
+ * 
+ * Returns #TRUE if the @tree_column can be resized by the end user.
+ * 
+ * Return value: #TRUE, if the @tree_column can be resized.
+ **/
 gboolean
 gtk_tree_view_column_get_resizable (GtkTreeViewColumn *tree_column)
 {
@@ -1528,9 +1597,10 @@ gtk_tree_view_column_get_width (GtkTreeViewColumn *tree_column)
  * @fixed_width: The size to set @tree_column to. Must be greater than 0.
  * 
  * Sets the size of the column in pixels.  This is meaningful only if the sizing
- * type is #GTK_TREE_VIEW_COLUMN_FIXED.  In this case, the value is discarded
- * as the size of the column is based on the calculated width of the column. The
- * width is clamped to the min/max width for the column.
+ * type is #GTK_TREE_VIEW_COLUMN_FIXED.  The size of the column is clamped to
+ * the min/max width for the column.  Please note that the min/max width of the
+ * column doesn't actually affect the "fixed_width" property of the widget, just
+ * the actual size when displayed.
  **/
 void
 gtk_tree_view_column_set_fixed_width (GtkTreeViewColumn *tree_column,
@@ -1539,9 +1609,6 @@ gtk_tree_view_column_set_fixed_width (GtkTreeViewColumn *tree_column,
   g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column));
   g_return_if_fail (fixed_width > 0);
 
-  if (tree_column->column_type == GTK_TREE_VIEW_COLUMN_AUTOSIZE)
-    return;
-
   tree_column->fixed_width = fixed_width;
 
   if (tree_column->tree_view &&
@@ -1550,6 +1617,8 @@ gtk_tree_view_column_set_fixed_width (GtkTreeViewColumn *tree_column,
     {
       gtk_widget_queue_resize (tree_column->tree_view);
     }
+
+  g_object_notify (G_OBJECT (tree_column), "fixed_width");
 }
 
 /**
@@ -1596,12 +1665,14 @@ gtk_tree_view_column_set_min_width (GtkTreeViewColumn *tree_column,
     }
 
   tree_column->min_width = min_width;
+  g_object_freeze_notify (G_OBJECT (tree_column));
   if (tree_column->max_width != -1 && tree_column->max_width < min_width)
     {
       tree_column->max_width = min_width;
       g_object_notify (G_OBJECT (tree_column), "max_width");
     }
   g_object_notify (G_OBJECT (tree_column), "min_width");
+  g_object_thaw_notify (G_OBJECT (tree_column));
 }
 
 /**
@@ -1650,12 +1721,14 @@ gtk_tree_view_column_set_max_width (GtkTreeViewColumn *tree_column,
     }
 
   tree_column->max_width = max_width;
+  g_object_freeze_notify (G_OBJECT (tree_column));
   if (max_width != -1 && max_width < tree_column->min_width)
     {
       tree_column->min_width = max_width;
       g_object_notify (G_OBJECT (tree_column), "min_width");
     }
   g_object_notify (G_OBJECT (tree_column), "max_width");
+  g_object_thaw_notify (G_OBJECT (tree_column));
 }
 
 /**
@@ -1805,11 +1878,10 @@ gtk_tree_view_column_set_widget (GtkTreeViewColumn *tree_column,
  * gtk_tree_view_column_get_widget:
  * @tree_column: A #GtkTreeViewColumn.
  * 
- * Returns the #GtkWidget in the button in the column header.  If a custom
- * widget has not been set, then this will be a #GtkAlignment with a #GtkLabel
- * in it.
+ * Returns the #GtkWidget in the button on the column header.  If a custom
+ * widget has not been set then %NULL is returned.
  * 
- * Return value: The #GtkWidget in the column header.
+ * Return value: The #GtkWidget in the column header, or %NULL
  **/
 GtkWidget *
 gtk_tree_view_column_get_widget (GtkTreeViewColumn *tree_column)
@@ -1861,6 +1933,14 @@ gtk_tree_view_column_get_alignment (GtkTreeViewColumn *tree_column)
   return tree_column->xalign;
 }
 
+/**
+ * gtk_tree_view_column_set_reorderable:
+ * @tree_column: A #GtkTreeViewColumn
+ * @reorderable: #TRUE, if the column can be reordered.
+ * 
+ * If @reorderable is #TRUE, then the column can be reordered by the end user
+ * dragging the header.
+ **/
 void
 gtk_tree_view_column_set_reorderable (GtkTreeViewColumn *tree_column,
 				      gboolean           reorderable)
@@ -1878,6 +1958,14 @@ gtk_tree_view_column_set_reorderable (GtkTreeViewColumn *tree_column,
   g_object_notify (G_OBJECT (tree_column), "reorderable");
 }
 
+/**
+ * gtk_tree_view_column_get_reorderable:
+ * @tree_column: A #GtkTreeViewColumn
+ * 
+ * Returns #TRUE if the @tree_column can be reordered by the user.
+ * 
+ * Return value: #TRUE if the @tree_column can be reordered by the user.
+ **/
 gboolean
 gtk_tree_view_column_get_reorderable (GtkTreeViewColumn *tree_column)
 {
@@ -2125,6 +2213,7 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
 {
   GList *list;
   gboolean first_cell = TRUE;
+  gint focus_line_width;
 
   g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column));
 
@@ -2133,6 +2222,8 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
   if (width)
     * width = 0;
 
+  gtk_widget_style_get (tree_column->tree_view, "focus-line-width", &focus_line_width, NULL);
+  
   for (list = tree_column->cell_list; list; list = list->next)
     {
       GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
@@ -2156,11 +2247,11 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
 				  &new_height);
 
       if (height)
-	* height = MAX (*height, new_height);
-      info->requested_width = MAX (info->requested_width, new_width);
+	* height = MAX (*height, new_height + focus_line_width * 2);
+      info->requested_width = MAX (info->requested_width, new_width + focus_line_width * 2);
       if (width)
 	* width += info->requested_width;
-      first_cell = TRUE;
+      first_cell = FALSE;
     }
 }
 
@@ -2168,15 +2259,24 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
  * code.  Rather than duplicate them, we put them together to keep the code in
  * one place
  */
-static void
-gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
-					   GdkWindow         *window,
-					   GdkRectangle      *background_area,
-					   GdkRectangle      *cell_area,
-					   GdkRectangle      *expose_area,
-					   guint              flags,
-					   gboolean           render,
-					   GdkRectangle      *focus_rectangle)
+enum {
+  CELL_ACTION_RENDER,
+  CELL_ACTION_FOCUS,
+  CELL_ACTION_EVENT
+};
+
+static gboolean
+gtk_tree_view_column_cell_process_action (GtkTreeViewColumn  *tree_column,
+					  GdkWindow          *window,
+					  GdkRectangle       *background_area,
+					  GdkRectangle       *cell_area,
+					  guint               flags,
+					  gint                action,
+					  GdkRectangle       *expose_area,     /* RENDER */
+					  GdkRectangle       *focus_rectangle, /* FOCUS  */
+					  GtkCellEditable   **editable_widget, /* EVENT  */
+					  GdkEvent           *event,           /* EVENT  */
+					  gchar              *path_string)     /* EVENT  */
 {
   GList *list;
   GdkRectangle real_cell_area;
@@ -2184,6 +2284,7 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
   gint full_requested_width = 0;
   gint extra_space;
   gint min_x, min_y, max_x, max_y;
+  gint focus_line_width;
 
   min_x = G_MAXINT;
   min_y = G_MAXINT;
@@ -2192,6 +2293,9 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 
   real_cell_area = *cell_area;
 
+  gtk_widget_style_get (GTK_WIDGET (tree_column->tree_view),
+			"focus-line-width", &focus_line_width,
+			NULL);
   /* Find out how my extra space we have to allocate */
   for (list = tree_column->cell_list; list; list = list->next)
     {
@@ -2221,7 +2325,8 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 
       real_cell_area.width = info->requested_width +
 	(info->expand?extra_space:0);
-      if (render)
+      real_cell_area.x += focus_line_width;
+      if (action == CELL_ACTION_RENDER)
 	{
 	  gtk_cell_renderer_render (info->cell,
 				    window,
@@ -2231,7 +2336,105 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 				    expose_area,
 				    flags);
 	}
-      else
+      else if (action == CELL_ACTION_FOCUS)
+	{
+	  gint x_offset, y_offset, width, height;
+
+	  gtk_cell_renderer_get_size (info->cell,
+				      tree_column->tree_view,
+				      &real_cell_area,
+				      &x_offset, &y_offset,
+				      &width, &height);
+
+	  if (min_x > (real_cell_area.x + x_offset))
+	    min_x = real_cell_area.x + x_offset;
+	  if (max_x < real_cell_area.x + x_offset + width)
+	    max_x = real_cell_area.x + x_offset + width;
+	  if (min_y > (real_cell_area.y + y_offset))
+	    min_y = real_cell_area.y + y_offset;
+	  if (max_y < real_cell_area.y + y_offset + height)
+	    max_y = real_cell_area.y + y_offset + height;
+	}
+      else if (action == CELL_ACTION_EVENT)
+	{
+	  gint x_offset, y_offset, width, height;
+	  gboolean try_event = FALSE;
+
+	  if (event)
+	    {
+	      if (real_cell_area.x <= ((GdkEventButton *)event)->x &&
+		  real_cell_area.x + real_cell_area.width > ((GdkEventButton *)event)->x)
+		try_event = TRUE;
+	    }
+	  else /* if (info->has_focus)*/
+	    /* FIXME 73676: allow focusing individual cells */
+	    {
+	      try_event = TRUE;
+	    }
+	  if (try_event)
+	    {
+	      gboolean visible, mode;
+
+	      g_object_get (G_OBJECT (info->cell),
+			    "visible", &visible,
+			    "mode", &mode,
+			    NULL);
+	      if (visible && mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
+		{
+		  if (gtk_cell_renderer_activate (info->cell,
+						  event,
+						  tree_column->tree_view,
+						  path_string,
+						  background_area,
+						  cell_area,
+						  flags))
+		    return TRUE;
+		}
+	      else if (visible && mode == GTK_CELL_RENDERER_MODE_EDITABLE)
+		{
+		  *editable_widget =
+		    gtk_cell_renderer_start_editing (info->cell,
+						     event,
+						     tree_column->tree_view,
+						     path_string,
+						     background_area,
+						     cell_area,
+						     flags);
+
+		  if (*editable_widget != NULL)
+		    {
+		      g_return_val_if_fail (GTK_IS_CELL_EDITABLE (*editable_widget), FALSE);
+		      
+		      return TRUE;
+		    }
+		}
+	    }
+	}
+      real_cell_area.x += (info->requested_width + tree_column->spacing);
+    }
+  for (list = g_list_last (tree_column->cell_list); list; list = list->prev)
+    {
+      GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
+
+      if (info->pack == GTK_PACK_START)
+	continue;
+
+      if (! info->cell->visible)
+	continue;
+
+      real_cell_area.width = info->requested_width +
+	(info->expand?extra_space:0);
+      if (action == CELL_ACTION_RENDER)
+	{
+	  gtk_cell_renderer_render (info->cell,
+				    window,
+				    tree_column->tree_view,
+				    background_area,
+				    &real_cell_area,
+				    expose_area,
+				    flags);
+	}
+      else if (action == CELL_ACTION_FOCUS)
 	{
 	  gint x_offset, y_offset, width, height;
 
@@ -2252,45 +2455,25 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 	}
       real_cell_area.x += (info->requested_width + tree_column->spacing);
     }
-  for (list = g_list_last (tree_column->cell_list); list; list = list->prev)
-    {
-      GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
 
-      if (info->pack == GTK_PACK_START)
-	continue;
-
-      if (! info->cell->visible)
-	continue;
-
-      real_cell_area.width = info->requested_width +
-	(info->expand?extra_space:0);
-      gtk_cell_renderer_render (info->cell,
-				window,
-				tree_column->tree_view,
-				background_area,
-				&real_cell_area,
-				expose_area,
-				flags);
-      real_cell_area.x += (info->requested_width + tree_column->spacing);
-    }
-  if (! render)
+  if (action == CELL_ACTION_FOCUS)
     {
       if (min_x >= max_x || min_y >= max_y)
 	{
 	  *focus_rectangle = *cell_area;
-	  focus_rectangle->x -= 1;
-	  focus_rectangle->y -= 1;
-	  focus_rectangle->width += 2;
-	  focus_rectangle->height += 2;
+	  /* don't change the focus_rectangle, just draw it nicely inside
+	   * the cell area */
 	}
       else
 	{
-	  focus_rectangle->x = min_x - 1;
-	  focus_rectangle->y = min_y - 1;
-	  focus_rectangle->width = (max_x - min_x) + 2;
-	  focus_rectangle->height = (max_y - min_y) + 2;
+	  focus_rectangle->x = min_x - focus_line_width;
+	  focus_rectangle->y = min_y - focus_line_width;
+	  focus_rectangle->width = (max_x - min_x) + 2 * focus_line_width;
+	  focus_rectangle->height = (max_y - min_y) + 2 * focus_line_width;
 	}
     }
+
+  return FALSE;
 }
 
 /**
@@ -2306,26 +2489,26 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
  * #GtkTreeView.
  **/
 void
-gtk_tree_view_column_cell_render (GtkTreeViewColumn *tree_column,
-				  GdkWindow         *window,
-				  GdkRectangle      *background_area,
-				  GdkRectangle      *cell_area,
-				  GdkRectangle      *expose_area,
-				  guint              flags)
+_gtk_tree_view_column_cell_render (GtkTreeViewColumn *tree_column,
+				   GdkWindow         *window,
+				   GdkRectangle      *background_area,
+				   GdkRectangle      *cell_area,
+				   GdkRectangle      *expose_area,
+				   guint              flags)
 {
   g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column));
   g_return_if_fail (background_area != NULL);
   g_return_if_fail (cell_area != NULL);
   g_return_if_fail (expose_area != NULL);
 
-  gtk_tree_view_column_cell_render_or_focus (tree_column,
-					     window,
-					     background_area,
-					     cell_area,
-					     expose_area,
-					     flags,
-					     TRUE,
-					     NULL);
+  gtk_tree_view_column_cell_process_action (tree_column,
+					    window,
+					    background_area,
+					    cell_area,
+					    flags,
+					    CELL_ACTION_RENDER,
+					    expose_area,
+					    NULL, NULL, NULL, NULL);
 }
 
 gboolean
@@ -2337,96 +2520,81 @@ _gtk_tree_view_column_cell_event (GtkTreeViewColumn  *tree_column,
 				  GdkRectangle       *cell_area,
 				  guint               flags)
 {
-  gboolean visible, mode;
-
   g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column), FALSE);
 
-  g_object_get (G_OBJECT (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell),
-		"visible", &visible,
-		"mode", &mode,
-		NULL);
-  if (visible && mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
-    {
-      if (gtk_cell_renderer_activate (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell,
-				      event,
-				      tree_column->tree_view,
-				      path_string,
-				      background_area,
-				      cell_area,
-				      flags))
-	return TRUE;
-    }
-  else if (visible && mode == GTK_CELL_RENDERER_MODE_EDITABLE)
-    {
-      *editable_widget = gtk_cell_renderer_start_editing (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell,
-							  event,
-							  tree_column->tree_view,
-							  path_string,
-							  background_area,
-							  cell_area,
-							  flags);
-
-      if (*editable_widget != NULL)
-	{
-	  g_return_val_if_fail (GTK_IS_CELL_EDITABLE (*editable_widget), FALSE);
-
-	  return TRUE;
-	}
-    }
-  return FALSE;
+  return gtk_tree_view_column_cell_process_action (tree_column,
+						   NULL,
+						   background_area,
+						   cell_area,
+						   flags,
+						   CELL_ACTION_EVENT,
+						   NULL, NULL,
+						   editable_widget,
+						   event,
+						   path_string);
 }
 
 
 gboolean
-gtk_tree_view_column_cell_focus (GtkTreeViewColumn *tree_column,
-				 gint               direction)
+_gtk_tree_view_column_cell_focus (GtkTreeViewColumn *tree_column,
+				  gint               direction)
 {
+  /* FIXME 73676: allow focusing individual cells */
   if (GTK_TREE_VIEW (tree_column->tree_view)->priv->focus_column == tree_column)
     return FALSE;
   return TRUE;
 }
 
 void
-gtk_tree_view_column_cell_draw_focus (GtkTreeViewColumn       *tree_column,
-				      GdkWindow               *window,
-				      GdkRectangle            *background_area,
-				      GdkRectangle            *cell_area,
-				      GdkRectangle            *expose_area,
-				      guint                    flags)
+_gtk_tree_view_column_cell_draw_focus (GtkTreeViewColumn       *tree_column,
+				       GdkWindow               *window,
+				       GdkRectangle            *background_area,
+				       GdkRectangle            *cell_area,
+				       GdkRectangle            *expose_area,
+				       guint                    flags)
 {
+  gint focus_line_width;
+  GtkStateType cell_state;
+  
   g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column));
+  gtk_widget_style_get (GTK_WIDGET (tree_column->tree_view),
+			"focus-line-width", &focus_line_width, NULL);
   if (tree_column->editable_widget)
     {
       /* This function is only called on the editable row when editing.
        */
-
+#if 0
       gtk_paint_focus (tree_column->tree_view->style,
 		       window,
 		       GTK_WIDGET_STATE (tree_column->tree_view),
 		       NULL,
 		       tree_column->tree_view,
 		       "treeview",
-		       cell_area->x - 1,
-		       cell_area->y - 1,
-		       cell_area->width + 2,
-		       cell_area->height + 2);
-      
+		       cell_area->x - focus_line_width,
+		       cell_area->y - focus_line_width,
+		       cell_area->width + 2 * focus_line_width,
+		       cell_area->height + 2 * focus_line_width);
+#endif      
     }
   else
     {
       GdkRectangle focus_rectangle;
-      gtk_tree_view_column_cell_render_or_focus (tree_column,
-						 window,
-						 background_area,
-						 cell_area,
-						 expose_area,
-						 flags,
-						 FALSE,
-						 &focus_rectangle);
-      
+      gtk_tree_view_column_cell_process_action (tree_column,
+						window,
+						background_area,
+						cell_area,
+						flags,
+						CELL_ACTION_FOCUS,
+						expose_area,
+						&focus_rectangle,
+						NULL, NULL, NULL);
+
+      cell_state = flags & GTK_CELL_RENDERER_SELECTED ? GTK_STATE_SELECTED :
+	      (flags & GTK_CELL_RENDERER_PRELIT ? GTK_STATE_PRELIGHT :
+	      (flags & GTK_CELL_RENDERER_INSENSITIVE ? GTK_STATE_INSENSITIVE : GTK_STATE_NORMAL));
       gtk_paint_focus (tree_column->tree_view->style,
 		       window,
-		       GTK_WIDGET_STATE (tree_column->tree_view),
+		       cell_state,
 		       NULL,
 		       tree_column->tree_view,
 		       "treeview",
@@ -2437,6 +2605,16 @@ gtk_tree_view_column_cell_draw_focus (GtkTreeViewColumn       *tree_column,
     }
 }
 
+/**
+ * gtk_tree_view_column_cell_is_visible:
+ * @tree_column: A #GtkTreeViewColumn
+ * 
+ * Returns #TRUE if any of the cells packed into the @tree_column are visible.
+ * For this to be meaningful, you must first initialize the cells with
+ * gtk_tree_view_column_cell_set_cell_data()
+ * 
+ * Return value: #TRUE, if any of the cells packed into the @tree_column are currently visible
+ **/
 gboolean
 gtk_tree_view_column_cell_is_visible (GtkTreeViewColumn *tree_column)
 {
@@ -2454,7 +2632,8 @@ gtk_tree_view_column_cell_is_visible (GtkTreeViewColumn *tree_column)
 }
 
 void
-gtk_tree_view_column_cell_set_dirty (GtkTreeViewColumn *tree_column)
+_gtk_tree_view_column_cell_set_dirty (GtkTreeViewColumn *tree_column,
+				      gboolean           install_handler)
 {
   GList *list;
 
@@ -2465,12 +2644,17 @@ gtk_tree_view_column_cell_set_dirty (GtkTreeViewColumn *tree_column)
       info->requested_width = 0;
     }
   tree_column->dirty = TRUE;
-  tree_column->requested_width = 0;
+  tree_column->resized_width = MAX (tree_column->requested_width, tree_column->button_request);
+  tree_column->requested_width = -1;
+  tree_column->width = 0;
 
   if (tree_column->tree_view &&
       GTK_WIDGET_REALIZED (tree_column->tree_view))
     {
-      _gtk_tree_view_install_mark_rows_col_dirty (GTK_TREE_VIEW (tree_column->tree_view));
+      if (install_handler)
+	_gtk_tree_view_install_mark_rows_col_dirty (GTK_TREE_VIEW (tree_column->tree_view));
+      else
+	GTK_TREE_VIEW (tree_column->tree_view)->priv->mark_rows_col_dirty = TRUE;
       gtk_widget_queue_resize (tree_column->tree_view);
     }
 }
