@@ -2536,7 +2536,7 @@ file_exists (const char *filename)
 void
 test_init ()
 {
-  if (file_exists ("../gdk-pixbuf/.libs/libpixbufloader-pnm.so"))
+  if (file_exists ("../gdk-pixbuf/libpixbufloader-pnm.la"))
     {
       putenv ("GDK_PIXBUF_MODULEDIR=../gdk-pixbuf/.libs");
       putenv ("GTK_IM_MODULE_FILE=../modules/input/gtk.immodules");
